// "newactual.h"
// Actual objects to be instantiated.

/*    Copyright (C) 1994,1995,1996  Steve Hardt

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 1, or (at your option)
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    This file is primarily the work of George Caswell.  As with all software
    in this version of Xevil, problems, questions, and other support
    information should be directed to him.  I encourage you also to take
    a look at the newest version of the "real" Xevil-  It is no longer
    freely source distributable, but it is still years ahead of this
    version, including more weapons, more character classes, moving platforms,
    and real net play.

    George Caswell
    timbuktu@wpi.edu
    http://adamant.res.wpi.net/~timbuktu/xevil

    For Xevil 2.0 and up:

    Steve Hardt
    satan@xevil.com
    http://www.xevil.com
*/


#ifndef NEWACTUAL_H
#define NEWACTUAL_H

#ifndef NO_PRAGMAS
#pragma interface
#endif

// Include Files
#include "actual.h"
#include "utils.h"
#include "coord.h"
#include "world.h"
#include "locator.h"
#include "physical.h"


class Bazooka: public Gun {
 public:
  Bazooka(WorldP w,LocatorP l,const Pos &pos);
  static Size get_size_max();
  static PhysicalP create(WorldP,LocatorP,const Pos &);

  const static GunContext context;
  /* NOTE: Making this public is a bit of a hack.  Locator::register_contexts()
     needs to get to it and I don't want to have to make a separate accessor
     function for every class. */

private:
  virtual Size get_shot_size(Dir);
  virtual PhysicalP create_shot(PhysicalP,WorldP,LocatorP,const Pos &,Dir dir);

  static GunXdata xdata;
};



// Bazooka-type rocket
class Rocket: public Shot {
 public:
  Rocket(WorldP w,LocatorP l,const Pos &pos,const Id &shooter,Dir dir);
  static Size get_size(Dir);
  static Size get_size_max();
  static PhysicalP create(WorldP,LocatorP,const Pos &);

  static const Stats &get_stats() {return stats;}

  virtual void die();

  virtual int get_anim_time();

  virtual void act();

  const static ShotContext context;
  /* NOTE: Making this public is a bit of a hack.  Locator::register_contexts()
     needs to get to it and I don't want to have to make a separate accessor
     function for every class. */

 private:
//  Id shooterId;
  Id shooter;

  Timer armingTimer;

  static ShotXdata xdata;
  static Stats stats;
};



class BloodSplat: public Falling {
public:
  BloodSplat(const FallingContext &fc, FallingXdata &fx, WorldP w,LocatorP l,const Pos &pos, Dir dir,Speed speed);
  BloodSplat(const FallingContext &fc, FallingXdata &fx, WorldP, LocatorP,const Pos &, const Vel &);

  virtual Boolean collidable();

  virtual void act();

 private:
  Timer timer;

  Boolean splat;
};


class Blood: public BloodSplat {
public:
  Blood(WorldP w, LocatorP l, const Pos &pos, Dir dir, Speed speed);
  Blood(WorldP, LocatorP,const Pos &, const Vel &);
  static Size get_size(Dir dir);
  static Size get_size_max();
  static PhysicalP create(WorldP,LocatorP,const Pos &);

  static const Stats &get_stats() {return stats;}

  const static FallingContext context;
  /* NOTE: Making this public is a bit of a hack.  Locator::register_contexts()
     needs to get to it and I don't want to have to make a separate accessor
     function for every class. */
private:
  static FallingXdata xdata;
  static Stats stats;
};


class OilBlood: public BloodSplat {
public:
  OilBlood(WorldP w,LocatorP l,const Pos &pos,
	  Dir dir,Speed speed);
  OilBlood(WorldP, LocatorP,const Pos &, const Vel &);
  static Size get_size(Dir dir);
  static Size get_size_max();
  static PhysicalP create(WorldP,LocatorP,const Pos &);

  static const Stats &get_stats() {return stats;}

  const static FallingContext context;
  /* NOTE: Making this public is a bit of a hack.  Locator::register_contexts()
     needs to get to it and I don't want to have to make a separate accessor
     function for every class. */

private:
  static FallingXdata xdata;
  static Stats stats;
};


class AlienBlood: public BloodSplat {
public:
  AlienBlood(WorldP w,LocatorP l,const Pos &pos,
	  Dir dir,Speed speed);
  AlienBlood(WorldP, LocatorP,const Pos &, const Vel &);
  static Size get_size(Dir dir);
  static Size get_size_max();
  static PhysicalP create(WorldP,LocatorP,const Pos &);

  static const Stats &get_stats() {return stats;}

  const static FallingContext context;
  /* NOTE: Making this public is a bit of a hack.  Locator::register_contexts()
     needs to get to it and I don't want to have to make a separate accessor
     function for every class. */

private:
  static FallingXdata xdata;
  static Stats stats;
};


class Smoke: public Moving {
 public:
  Smoke(WorldP w,LocatorP l,const Pos &pos);
  static Size get_size(); // Doubles as get_size_max().
  static PhysicalP create(WorldP,LocatorP,const Pos &);

  virtual Boolean collidable();
  virtual Boolean ignore_walls();

  virtual int get_drawing_level();

  virtual void act();

  const static MovingContext context;

 private:
  Timer timer;
  static MovingXdata xdata;

  virtual int get_anim_time();
  Pos uncenter(Pos mid);
};



class Doppeloid: public AutoUse {
  // Not a AutoUse because Lemmings and other funny things get doppelganged.
 public:
  Doppeloid(const AutoUseContext &, AutoUseXdata &, WorldP,LocatorP,const
            Pos &);
  //Not to be called directly.
  
  virtual void use(PhysicalP);

 private:
  virtual PhysicalP create_physical(const Area &,PhysicalP original);
  /* EFFECTS: Create a new Physical which is of the same class as original. */
};


class Doppel: public Doppeloid {
  // Not a AutoUse because Lemmings and other funny things get doppelganged.
 public:
  Doppel(WorldP,LocatorP,const Pos &);
  static Size get_size_max();
  static PhysicalP create(WorldP,LocatorP,const Pos &);
  
  virtual void use(PhysicalP);

  static const Stats &get_stats() {return stats;}

  const static AutoUseContext context;

 private:

  static AutoUseXdata xdata;
  static Stats stats;
};


class TransDoppel: public Doppeloid {
 public:
  TransDoppel(WorldP,LocatorP,const Pos &);
  static Size get_size_max();
  static PhysicalP create(WorldP,LocatorP,const Pos &);
  
  virtual void use(PhysicalP);

  static const Stats &get_stats() {return stats;}

  const static AutoUseContext context;

 private:
  virtual PhysicalP create_physical(const Area &,PhysicalP original);
  /* EFFECTS: Create a new Physical which is of the same class as original. */
  static Boolean transmogify_target_filter(const PhysicalContext *);

  static AutoUseXdata xdata;
  static Stats stats;
};


/*************SCUD*************/
/*(Apologies to Rob Scrab.  Wonder if I make the list of nobholes for this.*/

class Scud: public virtual Creature, public Walking, public Fighter,
public User
{
 public:
  Scud(WorldP w,LocatorP l,const Pos &pos);
  static Size get_size_max();
  static PhysicalP create(WorldP,LocatorP,const Pos &);
  
  virtual Boolean is_fighter();
  virtual Boolean is_walking();
  virtual Boolean is_user();

  virtual void get_pixmap_mask(int dpyNum,Pixmap &,Pixmap &,Dir,int);
  virtual void get_size_offset_next(Size &size,Size &offset,Dir dirNext);

  virtual int get_weapons_num();
  virtual int get_items_num(); 
  virtual PhysicalP get_weapon(int);
  virtual PhysicalP get_item(int); 
  virtual PhysicalP get_weapon_current();
  virtual PhysicalP get_item_current();
  virtual int get_item_count();
  virtual void set_mapped_next(Boolean val);

  virtual ClassId get_weapon_string(const char *&str);
  
  virtual void act();

  virtual void update();
  
  virtual void die();

  virtual void collide(PhysicalP other);
  
  static const Stats &get_stats() {return stats;}

  const static CreatureContext creatureContext;
  /* NOTE: Making this public is a bit of a hack.  Locator::register_contexts()
     needs to get to it and I don't want to have to make a separate accessor
     function for every class. */


private:
  virtual void init_x(Xvars &);

//  Boolean disposable;  //If TRUE, will explode once it kills something.

  static CreatureXdata creatureXdata;
  static WalkingXdata walkingXdata;
  static FighterXdata fighterXdata;
  static UserXdata userXdata;
  
  const static WalkingContext walkingContext;
  const static FighterContext fighterContext;
  const static UserContext userContext;
  static Stats stats;
};


class ScudCounter {
public:
  ScudCounter();
  void init();
  int assign_number();
  int peek_number();
private:
  const int numbermax=2000;
  int numberused;
  int numberpool[numbermax];
};


class Vendor: public AutoUse {
   //Vending machine.  Use it and get a pet scud!
 public:
  Vendor(WorldP,LocatorP,const Pos &);
  static Size get_size_max();
  static PhysicalP create(WorldP,LocatorP,const Pos &);
  
  virtual void use(PhysicalP);

  static const Stats &get_stats() {return stats;}

  const static AutoUseContext context;

//  Boolean can_take() {return False;};

 private:
  UserP create_scud(const Area &,PhysicalP original);
  /* EFFECTS: Create a new Physical of Scud class. */

  static AutoUseXdata xdata;
  static Stats stats;
  static ScudCounter scudCounter;
};

/*********END SCUD*************/


class HumanTorch: public Shield {
//Turns the user into the human torch, flaming and heat-protected..
public:
   HumanTorch(WorldP w, LocatorP l, const Pos &pos);
   static Size get_size_max();
   static PhysicalP create(WorldP,LocatorP,const Pos &);
  
   virtual void use(PhysicalP);
   virtual ProtectionP create_protection(const Area&);

   static const Stats &get_stats() {return stats;}

   const static ShieldContext context;
private:
  static ShieldXdata xdata;
  static Stats stats;
};

class FlameOn: public Protection {
public:
   FlameOn(WorldP w, LocatorP l, const Area &area);
   static Size get_size_max();
   static PhysicalP create(WorldP,LocatorP,const Pos &);
   
   virtual void draw(Drawable, Xvars &, int, const Area &);
   virtual void act();

   const static ProtectionContext context;

private:
   virtual int corporeal_defend(int damage);
   virtual int heat_defend(int heat, Boolean secondary);
   
//   Area &flamearea;
   static ProtectionXdata xdata;
};


class Fire: public NewAnimated {
public:
  Fire(WorldP w,LocatorP l,const Pos &middle,Boolean coll = True);
  /* NOTE: If coll is False, the Fire will not collide with anything.
     I.e. Just for decoration. */

  static Size get_size_max();
  static PhysicalP create(WorldP,LocatorP,const Pos &);
  
  static int get_damage();

  virtual int get_drawing_level();

  virtual Boolean collidable();
  virtual Boolean ignore_walls();

  virtual void avoid(PhysicalP other);
  virtual void collide(PhysicalP other);
  virtual void corporeal_attack(PhysicalP,int);
  virtual void heat_attack(PhysicalP,int,Boolean);

  virtual void act();

  const static NewAnimatedContext context;

private:
  Boolean isCollidable;
  void makeSmoke();
  Pos uncenter(Pos mid);

  Timer smokeTimer;
  Frame frame;

  static NewAnimatedXdata xdata;
};



class FireBomb: public Falling {
public:
  FireBomb(WorldP w, LocatorP l, const Pos &pos, const Id &sh, Dir dir, Speed speed);
  FireBomb(WorldP w, LocatorP l, const Pos &pos, const Id &sh, const Vel &vel);

  static Size get_size(Dir dir);
  static Size get_size_max();
  static PhysicalP create(WorldP, LocatorP, const Pos &);

  virtual void set_quiet_death();
  virtual void act();
  virtual void die();

  static const Stats &get_stats() {return stats;}

  const static FallingContext context;


private:
  Timer timer;
  Id shooter;
  Boolean defused;

  static FallingXdata xdata;
  static Stats stats;
};


class FireMaker: public Physical {
public:
  FireMaker(Id sh, WorldP w, LocatorP l, const Pos &p);

  virtual void draw(Drawable, Xvars &, int, const Area &);
  virtual Boolean collidable();

  virtual const Area &get_area();
  virtual const Area &get_area_next();
  virtual void act();

  static PhysicalP create(WorldP, LocatorP, const Pos &);

  static Size get_size_max();

  const static PhysicalContext context;

private:
  Id shooterId;
  Pos pos;
  Area area;
  Timer timer;

  const int num_to_make=10;
};


class FireBombs: public Gun {
public:
  FireBombs(WorldP w, LocatorP l, const Pos &p);

  static Size get_size_max();
  static PhysicalP create(WorldP, LocatorP, const Pos &);

  const static GunContext context;

private:
  virtual Size get_shot_size(Dir);
  virtual PhysicalP create_shot(PhysicalP, WorldP, LocatorP, const Pos &, Dir dir);

  static GunXdata xdata;
};



class RockMan: public virtual Creature, public Walking, public User,
public BuiltIn
{
public:
  RockMan(WorldP,LocatorP,const Pos &);
  static Size get_size_max();
  static PhysicalP create(WorldP,LocatorP,const Pos &);

  virtual Boolean is_walking();
  virtual Boolean is_user();
  virtual Boolean is_built_in();
  
  virtual int get_weapons_num();
  virtual int get_items_num(); 
  virtual PhysicalP get_weapon(int);
  virtual PhysicalP get_item(int); 
  virtual PhysicalP get_weapon_current();
  virtual PhysicalP get_item_current();
  virtual int get_item_count();
  virtual void set_mapped_next(Boolean val);

  virtual ClassId get_weapon_string(const char *&str);

  virtual void act();

  virtual void update();
  
  virtual void die();

  virtual void collide(PhysicalP other);

  static const Stats &get_stats() {return stats;}
  
  const static CreatureContext creatureContext;
  /* NOTE: Making this public is a bit of a hack.  Locator::register_contexts()
     needs to get to it and I don't want to have to make a separate accessor
     function for every class. */


#ifndef PROTECTED_IS_PUBLIC
 protected:
#endif
  virtual int get_anim_time();

  virtual Dir compute_weapon_dir(ITcommand);

  virtual Size get_shot_size(Dir);

  virtual PhysicalP create_shot(const Pos &,Dir);

  
private:
  static CreatureXdata creatureXdata;
  static WalkingXdata walkingXdata;
  static BuiltInXdata builtInXdata;
  static UserXdata userXdata;
  
  const static WalkingContext walkingContext;
  const static BuiltInContext builtInContext;
  const static UserContext userContext;
  static Stats stats;
};


class Buster: public Shot {
 public:
  Buster(WorldP w,LocatorP l,const Pos &pos,const Id &shooter,
	Dir dir);
  static Size get_size(Dir);
  static Size get_size_max();
  static PhysicalP create(WorldP,LocatorP,const Pos &);

  virtual void heat_attack(PhysicalP,int,Boolean);

  static Dir compute_weapon_dir(ITcommand);
  /* EFFECTS: Weapons firing a Laser shoot diagonal shots different than
     most weapons. */

  static const Stats &get_stats() {return stats;}

  const static ShotContext context;
  /* NOTE: Making this public is a bit of a hack.  Locator::register_contexts()
     needs to get to it and I don't want to have to make a separate accessor
     function for every class. */


 private:
  static ShotXdata xdata;
  static Stats stats;
};



class RockDeath: public Moving {
 public:
  RockDeath(WorldP w,LocatorP l,const Pos &pos, int time);
  static Size get_size(); // Doubles as get_size_max().
  static PhysicalP create(WorldP,LocatorP,const Pos &);

  virtual Boolean collidable();
  virtual Boolean ignore_walls();

  virtual int get_drawing_level();

  virtual void act();

  const static MovingContext context;

 private:
  Timer timer;
  static MovingXdata xdata;

  virtual int get_anim_time();
  Pos uncenter(Pos mid);
};



class RockEnter: public NewAnimated {
 public:
  RockEnter(WorldP w,LocatorP l,const Pos &pos);
  static Size get_size(); // Doubles as get_size_max().
  static PhysicalP create(WorldP,LocatorP,const Pos &);

  virtual Boolean collidable();
  virtual Boolean ignore_walls();
  virtual Boolean is_creature() {return True;}
  //To avoid confusing those poor machine AI's, bless their weary souls..

  virtual int get_drawing_level();

  virtual void act();

  const static NewAnimatedContext context;

 private:
  static NewAnimatedXdata xdata;
  Boolean onTarget;
  Frame frame;

  Pos relocate(Pos mid);
  Pos target;
};



class KShield: public Shield {
public:
  KShield(WorldP,LocatorP,const Pos &);
  static Size get_size_max();
  static PhysicalP create(WorldP,LocatorP,const Pos &);

  virtual void use(PhysicalP);

  static const Stats &get_stats() {return stats;}

  const static ShieldContext context;
  /* NOTE: Making this public is a bit of a hack.  Locator::register_contexts()
     needs to get to it and I don't want to have to make a separate accessor
     function for every class. */

  
private:
  virtual ProtectionP create_protection(const Area &);

  static ShieldXdata xdata;
  static Stats stats;
};



class RShield: public Shield {
public:
  RShield(WorldP,LocatorP,const Pos &);
  static Size get_size_max();
  static PhysicalP create(WorldP,LocatorP,const Pos &);

  virtual void use(PhysicalP);

  static const Stats &get_stats() {return stats;}

  const static ShieldContext context;
  /* NOTE: Making this public is a bit of a hack.  Locator::register_contexts()
     needs to get to it and I don't want to have to make a separate accessor
     function for every class. */

  
private:
  virtual ProtectionP create_protection(const Area &);

  static ShieldXdata xdata;
  static Stats stats;
};




// Protects from K_PROTECTION_K damage in corporeal attacks,
// Maximum of K_THRESHOLD damage on each attack.
class KProtection: public Protection {
public:
  KProtection(WorldP,LocatorP,const Area &);
  static Size get_size_max();
  static PhysicalP create(WorldP,LocatorP,const Pos &);

  const static ProtectionContext context;
  /* NOTE: Making this public is a bit of a hack.  Locator::register_contexts()
     needs to get to it and I don't want to have to make a separate accessor
     function for every class. */


private:
  virtual int corporeal_defend(int);
  virtual int heat_defend(int heat,Boolean);

  int k; // Number of damage that it can still take.
  static ProtectionXdata xdata;
};



// Protects user from R_PROTECTION_PERCENT percent of the damage from each
// physical or heat attack.  Unlimited time.
class RProtection: public Protection {
public:
  RProtection(WorldP,LocatorP,const Area &);
  static Size get_size_max();
  static PhysicalP create(WorldP,LocatorP,const Pos &);

  const static ProtectionContext context;
  /* NOTE: Making this public is a bit of a hack.  Locator::register_contexts()
     needs to get to it and I don't want to have to make a separate accessor
     function for every class. */


private:
  virtual int corporeal_defend(int);
  virtual int heat_defend(int heat,Boolean);

  int r, r_h;

  Timer timer;
  static ProtectionXdata xdata;
};



class JetPack: public AutoUse {
public:
   JetPack(WorldP w, LocatorP l, const Pos &pos);
   static Size get_size_max();
   static PhysicalP create(WorldP,LocatorP,const Pos &);
  
   virtual void use(PhysicalP);

   static const Stats &get_stats() {return stats;}

   const static AutoUseContext context;
private:
  static AutoUseXdata xdata;
  static Stats stats;
};



class Kenny: public virtual Creature, public Grounded, public Suicide {
public:
   Kenny(WorldP w, LocatorP l, const Pos &rawPos);

   static Size get_size_max();
   static PhysicalP create(WorldP, LocatorP, const Pos &);

   virtual Boolean is_suicide();
   virtual void act();

   virtual void die();    //How could we define Kenny without this??

   const static CreatureContext creatureContext;

private:
   const static GroundedContext groundedContext;
   const static SuicideContext suicideContext;
   static CreatureXdata creatureXdata;
   static GroundedXdata groundedXdata;
   static SuicideXdata suicideXdata;
   static Stats stats;
};



class SwastikaRang: public Shot {
 public:
  SwastikaRang(WorldP w,LocatorP l,const Pos &pos,const Id &shooter,Dir dir);
  static Size get_size();
  static PhysicalP create(WorldP,LocatorP,const Pos &);

  static const Stats &get_stats() {return stats;}

  virtual Boolean die_on_impact();

  virtual void act();
  virtual void collide(PhysicalP other);

  const static ShotContext context;
  /* NOTE: Making this public is a bit of a hack.  Locator::register_contexts()
     needs to get to it and I don't want to have to make a separate accessor
     function for every class. */

 private:
  Id shooter;

  Boolean returning;
  Boolean returningbegun;
  Timer returnTimer;

  static ShotXdata xdata;
  static Stats stats;
};



class SwastikaRangs: public Gun {
 public:
  SwastikaRangs(WorldP w,LocatorP l,const Pos &pos);
  static Size get_size_max();
  static PhysicalP create(WorldP,LocatorP,const Pos &);

  const static GunContext context;

 private:
  virtual Size get_shot_size(Dir);
  virtual PhysicalP create_shot(PhysicalP,WorldP,LocatorP,const Pos &,Dir dir);

  static GunXdata xdata;
};



#endif

